 /** p3dExecTaskCodels.c
 **
 ** Codels appeles par la tache p3dExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

/* #ifndef UNIX */
/* #include <vxWorks.h> */
/* #include <timers.h> */
/* #else */
#include <portLib.h>
/* #endif */

#ifndef HAS_POSIX_CLOCK
#include <sys/time.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "server/p3dHeader.h"

#include <t3d/t3d.h>
#include <libp3d.h>

#include "demP3dPosterToP3dMap.h"
#if defined(ATRV) || defined(ATRV_JUNIOR)
#include "rflexConst.h"
#elif defined(RMP400)
#include "rmp400Const.h"
#endif


#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

/*---- macros ---------*/
#ifndef RAD_TO_DEG
#define RAD_TO_DEG(x) ( (x)*180.0/M_PI )
#endif

#ifndef DEG_TO_RAD
#define DEG_TO_RAD(x) ( (x) * M_PI / 180.0)
#endif


/* To debug */
#define DEBUG 0

/* put VERBOSE to 1 if you want a lot of logs (e.g. cf configMonitor) */
#define VERBOSE 1

/* local variables */
static P3DPlanner *planner;	/* the beast */

static int pomMainFrameId;

static T3D goalToOrigin;

static POSTER_ID pomPosterId;

/* var for map from dem */
static POSTER_ID demP3dPosterId;

static P3DPlanSolution* sol; /* Solution found by p3d */

static P3D_MAP *map;  /* The map */

/*--- predicted robot config. on current location ---*/
static P3DRobotConfig  predRobotConfig;

/* Local Functions */
short p3dReadDemMap (char* demP3dPosterName,
		     P3D_MAP  *p3dMap );


/* file pointer for the monitor log */
static FILE*  file;


/* Function for arc dump */
static P3DArcDump * ArcsDump;	
static void p3dDumpToFile(T3D* m2o, double, double, double);
static int p3dDumpEvaluateNbArcs(int n, int depth);

/*
 * p3dExecTaskInit ------------------------------------------------------
 *
 * Codel d'initialisation (SDI/f, ...)
 *
 * Retourne:    OK ou ERROR
 */

STATUS
p3dExecTaskInit(int *bilan)
{
   double len[3], rad[2], arm[2];
   int max_nb_arcs;

   planner = p3dCreatePlanner();
   if (planner == NULL) {
      *bilan = errnoGet();
      return ERROR;
   }

   /* wait first planning request to get it */
   pomMainFrameId = -1;

   len[0] = SDI_F->geomP.len0 = P3D_DEFAULT_WHEEL_LEN0;
   len[1] = SDI_F->geomP.len1 = P3D_DEFAULT_WHEEL_LEN1;
   len[2] = SDI_F->geomP.len2 = P3D_DEFAULT_WHEEL_LEN2;
   
   rad[0] = SDI_F->geomP.rad0 = P3D_DEFAULT_WHEEL_RAD0;
   rad[1] = SDI_F->geomP.rad1 = P3D_DEFAULT_WHEEL_RAD1;

   arm[0] = SDI_F->geomP.arm0 = P3D_DEFAULT_WHEEL_ARM0;
   arm[1] = SDI_F->geomP.arm1 = P3D_DEFAULT_WHEEL_ARM1;

   SDI_F->geomP.scale = P3D_DEFAULT_WHEEL_SCALE;
   SDI_F->geomP.rollStep = DEG_TO_RAD(P3D_DEFAULT_WHEEL_ROLLSTEP);
   SDI_F->geomP.vMin = P3D_DEFAULT_VMIN;
   SDI_F->geomP.vMax = P3D_DEFAULT_VMAX;
   SDI_F->geomP.accel = P3D_DEFAULT_ACCEL;
   SDI_F->geomP.decel = P3D_DEFAULT_DECEL;
   SDI_F->geomP.wMax = P3D_DEFAULT_WMAX;

   /*-- New: Body params --- */
   SDI_F->geomP.bodyLength  = P3D_DEFAULT_BODY_LENGTH;
   SDI_F->geomP.bodyWidth   = P3D_DEFAULT_BODY_WIDTH;
   SDI_F->geomP.zBody       = P3D_DEFAULT_BODY_Z;
   SDI_F->geomP.groundClear = P3D_DEFAULT_GROUND_CLEARANCE;

   SDI_F->limitsP.rollLimits[0] = DEG_TO_RAD(P3D_DEFAULT_ROLLLIMIT0);
   SDI_F->limitsP.rollLimits[1] = DEG_TO_RAD(P3D_DEFAULT_ROLLLIMIT1);
   SDI_F->limitsP.rollLimits[2] = DEG_TO_RAD(P3D_DEFAULT_ROLLLIMIT2);

   SDI_F->limitsP.pitchLimits[0] = DEG_TO_RAD(P3D_DEFAULT_PITCHLIMIT0);
   SDI_F->limitsP.pitchLimits[1] = DEG_TO_RAD(P3D_DEFAULT_PITCHLIMIT1);
   SDI_F->limitsP.pitchLimits[2] = DEG_TO_RAD(P3D_DEFAULT_PITCHLIMIT2);

   SDI_F->limitsP.alphaLimits[0] = DEG_TO_RAD(P3D_DEFAULT_ALPHALIMIT0);
   SDI_F->limitsP.alphaLimits[1] = DEG_TO_RAD(P3D_DEFAULT_ALPHALIMIT1);
   SDI_F->limitsP.alphaLimits[2] = DEG_TO_RAD(P3D_DEFAULT_ALPHALIMIT2);

   SDI_F->limitsP.betaLimits[0] = DEG_TO_RAD(P3D_DEFAULT_BETALIMIT0);
   SDI_F->limitsP.betaLimits[1] = DEG_TO_RAD(P3D_DEFAULT_BETALIMIT1);
   SDI_F->limitsP.betaLimits[2] = DEG_TO_RAD(P3D_DEFAULT_BETALIMIT2);

   SDI_F->limitsP.unkMax = P3D_DEFAULT_UNKMAX;

   SDI_F->arcsP.nArcs = P3D_DEFAULT_ARC_N;
   SDI_F->arcsP.nNodes = P3D_DEFAULT_ARC_NNODES;
   SDI_F->arcsP.depth = P3D_DEFAULT_ARC_DEPTH;
   SDI_F->arcsP.rMin = P3D_DEFAULT_ARC_RMIN;
   SDI_F->arcsP.dMin = P3D_DEFAULT_ARC_DMIN;
   SDI_F->arcsP.goalToal = P3D_DEFAULT_ARC_GOALTOL;
   SDI_F->arcsP.p3d_dMin = P3D_DEFAULT_P3D_ARC_DMIN;

   SDI_F->planP.period = P3D_DEFAULT_PLAN_PERIOD;
   SDI_F->planP.thetaUTurnHi = P3D_DEFAULT_THETAUTURNHI;
   SDI_F->planP.thetaUTurnLow = P3D_DEFAULT_THETAUTURNLOW;

#if defined(ATRV) || defined(ATRV_JUNIOR) || defined(RMP400)
   SDI_F->speedRef.numRef = 0;
   SDI_F->speedRef.updatePeriod = 0;
   SDI_F->speedRef.v = SDI_F->speedRef.w = 0.;
   SDI_F->speedRef.vt = 0.;
   SDI_F->speedRef.vmax = P3D_DEFAULT_VMAX;
   SDI_F->speedRef.wmax = P3D_DEFAULT_WMAX;
   SDI_F->speedRef.linAccelMax = 1.; /* unused anyway */
   SDI_F->speedRef.angAccelMax = 1.;
#else
   SDI_F->speedRef.dir = LLOCO_FORWARD;
   SDI_F->speedRef.num = 0;
   SDI_F->speedRef.servoMode = LLOCO_SERVO_SPEED;
   SDI_F->speedRef.updatePeriod = 0;
   SDI_F->speedRef.x = SDI_F->speedRef.y = SDI_F->speedRef.theta = 0.;
   SDI_F->speedRef.v = SDI_F->speedRef.w = 0.;
#endif

   SDI_F->wpList.number = 0;

   fputs("p3d: configuring geometry\n", stderr);
   if (p3dConfigureGeom(planner,
			len, rad, arm,
			SDI_F->geomP.bodyLength,
			SDI_F->geomP.bodyWidth,
			SDI_F->geomP.zBody,
			SDI_F->geomP.groundClear,
			SDI_F->limitsP.rollLimits,
			SDI_F->limitsP.pitchLimits,
			SDI_F->limitsP.alphaLimits,
			SDI_F->limitsP.betaLimits,
			SDI_F->limitsP.unkMax)) {
      fputs("p3d: default geometry failed\n", stderr);
      fputs("p3d: set geometry manually\n", stderr);
   }

   fputs("p3d: configuring robot\n", stderr);
   if (p3dConfigureRobot(planner,
			   SDI_F->geomP.scale,
			   SDI_F->geomP.rollStep,
			   SDI_F->geomP.vMin,
			   SDI_F->geomP.vMax,
			   SDI_F->geomP.wMax,
			   SDI_F->geomP.accel,
			   SDI_F->geomP.decel )) {
      fputs("p3d: default bitmap generation failed\n", stderr);
      fputs("p3d: set geometry manually\n", stderr);
   }

   fputs("p3d: configuring arcs\n", stderr);
   if (p3dConfigureArcs(planner,
			P3D_DEFAULT_ARC_N,
			P3D_DEFAULT_ARC_NNODES,
			P3D_DEFAULT_ARC_DEPTH,
			P3D_DEFAULT_ARC_RMIN,
			P3D_DEFAULT_ARC_GOALTOL, 
			P3D_DEFAULT_P3D_ARC_DMIN)) {
      fputs("p3d: default arcs configuration failed\n", stderr);
      fputs("p3d: configure arcs manually\n", stderr);
   }

   sol = create_P3DPlanSolution(P3D_DEFAULT_ARC_DEPTH, P3D_DEFAULT_ARC_NNODES);

   max_nb_arcs = p3dDumpEvaluateNbArcs(P3D_DEFAULT_ARC_N, P3D_DEFAULT_ARC_DEPTH);
   ArcsDump = malloc(max_nb_arcs * sizeof(*ArcsDump));

   /* create the solution set of arcs. It will be used to convert P3D's solution
    * into the CurrentArc poster */

   /****************************/
   /** p3d Map allocation, with initialisations **/
   fputs("p3d: creating p3d map\n", stderr);
   map = create_p3dmap();

   
   /* Default Read P3dPoster Params */
   strncpy(SDI_F->demP3dPosterName.name, P3D_DEFAULT_DTM_POSTER_NAME, P3D_MAX_STRLEN);


   return OK;
}

/*
 * p3dExecTaskEnd -------------------------------------------------------
 *
 * Codel de terminaison
 * 
 * Retourne:    OK ou ERROR
 */

STATUS
p3dExecTaskEnd(void)
{
   if (planner != NULL) p3dDestroyPlanner(planner);
   planner = NULL;

   if (sol != NULL) destroy_P3DPlanSolution(sol);
   sol = NULL;

   if (ArcsDump != NULL)  free(ArcsDump);
   ArcsDump = NULL;

   return OK;
}
/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 */

/* p3dInitExec  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dInitExec(P3D_INIT *initParams, int *report)
{
    if (posterFind(initParams->pomPoster, &pomPosterId) == ERROR) {
	fprintf(stderr, "p3d: cannot find pom poster\n");
	*report = S_p3d_POSTER_NOT_FOUND;
	return ETHER;
    }
    return ETHER;
}

/*
 * SetLamaGeom ----------------------------------------------------------
 *
 * Bilans:      OK 
 *              S_p3d_EINVAL
 */

/* p3dSetLamaGeomExec  -  codel EXEC de SetLamaGeom
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dSetLamaGeomExec(P3DLamaGeom *geomP, int *bilan)
{
   double len[3], rad[2], arm[2];

   len[0] = geomP->len0;
   len[1] = geomP->len1;
   len[2] = geomP->len2;
   
   rad[0] = geomP->rad0;
   rad[1] = geomP->rad1;

   arm[0] = geomP->arm0;
   arm[1] = geomP->arm1;

   SDI_F->geomP.rollStep = DEG_TO_RAD(SDI_F->geomP.rollStep);

   if (p3dConfigureGeom(planner,
			len, rad, arm,
			geomP->bodyLength, geomP->bodyWidth, 
			geomP->zBody, geomP->groundClear,
			SDI_F->limitsP.rollLimits,
			SDI_F->limitsP.pitchLimits,
			SDI_F->limitsP.alphaLimits,
			SDI_F->limitsP.betaLimits,
			SDI_F->limitsP.unkMax)) {
      *bilan = errnoGet();
      return ETHER;
   }

   if (p3dConfigureRobot(planner,
			   SDI_F->geomP.scale,
			   SDI_F->geomP.rollStep,
			   SDI_F->geomP.vMin,
			   SDI_F->geomP.vMax,
			   SDI_F->geomP.wMax,
			   SDI_F->geomP.accel,
			   SDI_F->geomP.decel )) {
      *bilan = errnoGet();
      return ETHER;
   }

   return ETHER;
}

/*
 * SetLamaLimits --------------------------------------------------------
 *
 * Bilans:      OK 
 *              S_p3d_EINVAL
 */

/* p3dSetLamaLimitsExec  -  codel EXEC de SetLamaLimits
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dSetLamaLimitsExec(P3DLamaLimits *limitsP, int *bilan)
{
   double len[3], rad[2], arm[2];
   int i;

   len[0] = SDI_F->geomP.len0;
   len[1] = SDI_F->geomP.len1;
   len[2] = SDI_F->geomP.len2;
   
   rad[0] = SDI_F->geomP.rad0;
   rad[1] = SDI_F->geomP.rad1;

   arm[0] = SDI_F->geomP.arm0;
   arm[1] = SDI_F->geomP.arm1;

   /* Convert limits from deg to rad */
   for (i = 0; i < 3; i++)
   {
	   limitsP->rollLimits[i] = DEG_TO_RAD(limitsP->rollLimits[i]);
	   limitsP->pitchLimits[i] = DEG_TO_RAD(limitsP->pitchLimits[i]);
	   limitsP->alphaLimits[i] = DEG_TO_RAD(limitsP->alphaLimits[i]);
	   limitsP->betaLimits[i] = DEG_TO_RAD(limitsP->betaLimits[i]);
   }

   if (p3dConfigureGeom(planner,
			len, rad, arm,
			SDI_F->geomP.bodyLength, 
			SDI_F->geomP.bodyWidth, 
			SDI_F->geomP.zBody,
			SDI_F->geomP.groundClear,
			limitsP->rollLimits,
			limitsP->pitchLimits,
			limitsP->alphaLimits,
			limitsP->betaLimits,
			limitsP->unkMax)) {
      *bilan = errnoGet();
      return ETHER;
   }

   if (p3dConfigureRobot(planner,
			   SDI_F->geomP.scale,
			   SDI_F->geomP.rollStep,
			   SDI_F->geomP.vMin,
			   SDI_F->geomP.vMax,
			   SDI_F->geomP.wMax,
			   SDI_F->geomP.accel,
			   SDI_F->geomP.decel )) {
      *bilan = errnoGet();
      return ETHER;
   }

   return ETHER;
}

/*
 * SetArcsParams --------------------------------------------------------
 *
 * Bilans:      OK 
 *              S_p3d_EINVAL
 */

/* p3dSetArcsParamsExec  -  codel EXEC de SetArcsParams
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dSetArcsParamsExec(P3DArcsParams *arcsP, int *bilan)
{
   int max_nb_arcs;

   if (p3dConfigureArcs(planner,
			arcsP->nArcs,
			arcsP->nNodes,
			arcsP->depth,
			arcsP->rMin,
			arcsP->goalToal,
			arcsP->p3d_dMin)) {
      *bilan = errnoGet();
      return ETHER;
   }

   if (sol) {
	   destroy_P3DPlanSolution(sol);
	   sol = NULL;
   }
   sol = create_P3DPlanSolution(arcsP->depth, arcsP->nNodes);

   if (ArcsDump) {
	   free(ArcsDump);
	   ArcsDump = NULL;
   };

   max_nb_arcs = p3dDumpEvaluateNbArcs(P3D_DEFAULT_ARC_N, P3D_DEFAULT_ARC_DEPTH);
   ArcsDump = malloc(max_nb_arcs * sizeof(*ArcsDump));

   return ETHER;
}




/* p3dReadDemP3dPosterExec  -  codel EXEC de ReadDemP3dPoster
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dReadDemP3dPosterExec (GEN_STRING64* demP3dPosterName, int *bilan)
{

  if(p3dReadDemMap(demP3dPosterName->name, map))
    {
      fputs("p3dReadDemP3dPosterExec: Cannot read dem poster \n", stderr);
      *bilan = S_p3d_ENOPOSTER;
      return ETHER;
    }

  return ETHER;
}




/*------------------------------------------------------------------------
 * Track
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_p3d_ENOPOSTER
 */

static POSTER_ID trajPosterId;

/* p3dTrackStart  -  codel START de Track
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dTrackStart(P3DSDIIo *trackP, int *report)
{
   if (SDI_F->wpList.number != 0) {
      fputs("p3d: cannot track when active waypoints\n", stderr);
      *report = S_p3d_ENOPOSTER;
      return ETHER;
   }

   SDI_F->trackRef = 0;
   SDI_F->movingStatus.blocked_cycles = 0;
   SDI_F->movingStatus.status = P3D_MOVING;
   if (p3dGENPOS_TRAJ_POINTSPosterFind(trackP->fileName, &trajPosterId) != OK) {
      *report = S_p3d_ENOPOSTER;
      return ETHER;
   }

   return EXEC;
}

/* p3dTrackExec  -  codel EXEC de Track
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dTrackExec(P3DSDIIo *trackP, int *report)
{
   int wpIdx;
   switch(SDI_F->movingStatus.status) {
       case P3D_MOVING: break;
       case P3D_GOAL_REACHED: break;
       case P3D_FAILED_POSTER: 
	    *report = S_p3d_ENOPOSTER;
	    return END; // this is never used
       case P3D_BLOCKED: 
	    *report = S_p3d_BLOCKED;
	    return END;
       case P3D_FAILED:
	    *report = S_p3d_NO_PATH;
	    return END;
   };

   GENPOS_TRAJ_POINTS currentGoals;
   if (p3dGENPOS_TRAJ_POINTSPosterRead(trajPosterId, &currentGoals) != OK) {
      *report = S_p3d_ENOPOSTER;
      return END;
   }

   /* Check if the poster changed */
   if (SDI_F->trackRef != currentGoals.numRef)
   {
       fprintf(stderr, "p3d: new path ref=%i nbpts=%i\n", currentGoals.numRef, currentGoals.nbPts);
       /* the current robot goal changed, we may not want to keep u-turning */
       SDI_F->trackRef = currentGoals.numRef;
       SDI_F->wpList.number = currentGoals.nbPts;
       for (wpIdx = 0; wpIdx < currentGoals.nbPts; ++wpIdx)
       {
	   SDI_F->wpList.wps[wpIdx].x = currentGoals.points[wpIdx].x;
	   SDI_F->wpList.wps[wpIdx].y = currentGoals.points[wpIdx].y;
	   SDI_F->wpList.wps[wpIdx].dir = GENPOS_FORWARD_MOTION;
       }
    }
       
   return EXEC;
}

/* p3dTrackEnd  -  codel END de Track
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dTrackEnd(P3DSDIIo *trackP, int *bilan)

{
   SDI_F->wpList.number = 0;
   return ETHER;
}




/* p3dSaveP3dMapExec  -  codel EXEC de SaveP3dMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dSaveP3dMapExec (P3D_SAVEMAP_PARAMS *p3dSaveMapParams, int *bilan)
{
  short  result;

  result = write_p3dmap(map, p3dSaveMapParams->fileName);

  if (result) {
    fprintf(stderr, "p3d: cannot save p3dmap in %s\n", p3dSaveMapParams->fileName);
    *bilan = S_p3d_ENOENT;
    return ETHER;
   }

  return ETHER;
}




/*solange for a nice goto*/

ACTIVITY_EVENT p3dGoToStart(P3DWayPoint *(in_wpP), int *bilan)
{
   SDI_F->movingStatus.status = P3D_MOVING;
   SDI_F->movingStatus.blocked_cycles = 0;
   SDI_F->wpList.number = 1;
   SDI_F->wpList.wps[0] = *in_wpP;

   fprintf(stderr, "p3dGoToStart: New Goal: x=%f, y=%f\n", SDI_F->wpList.wps[0].x, SDI_F->wpList.wps[0].y);
   
   return EXEC;
}

ACTIVITY_EVENT p3dGoToEnd(P3DWayPoint *(in_wpP), int *bilan)
{
  SDI_F->wpList.number = 0;
  return ETHER;
}

ACTIVITY_EVENT p3dGoToFail(P3DWayPoint *(in_wpP), int *bilan)
{
  SDI_F->wpList.number = 0;
  return ETHER;
}

ACTIVITY_EVENT p3dGoToInter(P3DWayPoint *(in_wpP), int *bilan)
{
  //  *bilan = S_p3d_INTERRUPTED;
  SDI_F->wpList.number = 0;
  return ETHER;
}

ACTIVITY_EVENT p3dGoToExec(P3DWayPoint *(in_wpP), int *bilan)
{

  if (SDI_F->movingStatus.status == P3D_MOVING) {
    return EXEC;
  }
  else if (SDI_F->movingStatus.status == P3D_GOAL_REACHED){
     printf("p3dGoToExec : goal reached\n");
    return END;
  }
  else if (SDI_F->movingStatus.status == P3D_BLOCKED){
    printf("p3dGoToExec : blocked\n");
    *bilan = S_p3d_BLOCKED;
    return END;
  }
  else { /*FAILED*/
    printf("p3dGoToExec : failed\n");
    *bilan = S_p3d_NO_PATH;
    return END;
  }

}


static 
void solutionToPoster()
{
   P3DSolution* arcsol = &(SDI_F->path);
   int arcIndex;
   int k;

   arcsol->depth = sol->depth;
   for (arcIndex = 0; arcIndex < arcsol->depth; ++arcIndex)
   {
       P3DArcGen* arc = &(arcsol->arcs[arcIndex]);
       P3DArc* bestArc = &sol->bestArc[sol->depth - 1 - arcIndex];

       arc->r = bestArc->r;
       arc->cost = bestArc->cost;
       arc->current = bestArc->current;
       arc->n = bestArc->n;

       for(k=0; k < arc->n; k++)
       {
	   arc->nodes[k].x = bestArc->nodes[k].x;
	   arc->nodes[k].y = bestArc->nodes[k].y;
	   arc->nodes[k].theta = bestArc->nodes[k].theta;

	   arc->nodes[k].conf.q = bestArc->nodes[k].conf.q;
	   arc->nodes[k].conf.x = bestArc->nodes[k].conf.x;
	   arc->nodes[k].conf.y = bestArc->nodes[k].conf.y;
	   arc->nodes[k].conf.z = bestArc->nodes[k].conf.z;
	   arc->nodes[k].conf.dzBodyGround = bestArc->nodes[k].conf.dzBodyGround;
	   arc->nodes[k].conf.roll = bestArc->nodes[k].conf.roll;
	   arc->nodes[k].conf.pitch = bestArc->nodes[k].conf.pitch;
	   arc->nodes[k].conf.alpha1 = bestArc->nodes[k].conf.alpha1;
	   arc->nodes[k].conf.alpha2 = bestArc->nodes[k].conf.alpha2;
	   arc->nodes[k].conf.beta3 = bestArc->nodes[k].conf.beta3;
	   arc->nodes[k].conf.pUnk = bestArc->nodes[k].conf.pUnk;

	   arc->nodes[k].cost = bestArc->nodes[k].cost;
       }
    }
}

/*
 * p3dExecTaskPerm ------------------------------------------------------
 *
 * Codel de l'activite' permanente
 *
 * Retourne:    OK ou ERROR
 */

STATUS
p3dExecTaskPerm(int *bilan)
{
   static T3D o2m, g2m;		/* goal to main */
#ifdef HAS_POSIX_CLOCK
   struct timespec tp0, tp1; /*  mesure du temps */
#else
   struct timeval tv0, tv1;
#endif /* HAS_POSIX_CLOCK */
   static int nPeriods;
   int i, result;
   POM_POS pos;
   double timeWait, theta;
   T3D mainToOrigin;
   double saved_v, saved_w, saved_average_time;

   if (--nPeriods > 0) {
      fprintf(stderr, "p3d: waiting %d periods\n", nPeriods);
      SDI_F->speedRef.numRef++;
      return OK;
   }
   if (SDI_F->planP.period <= 0.0) {
      SDI_F->movingStatus.status = P3D_FAILED;
      goto resetRef;
   }
   if (SDI_F->wpList.number <= 0) { 
      goto resetRef;
   }

   /* Read Dem map and fill p3dMap with its info */
   if(p3dReadDemMap(SDI_F->demP3dPosterName.name, map) )
   {
      fprintf(stderr, "p3d: Reading demPoster failed\n");
      fprintf(stderr, "p3d: \t You should read it manually once\n");
      goto resetRef;
   }

   /* read robot position */
   if (p3dPOM_POSPosterRead(pomPosterId, &(pos)) == ERROR) {
      fprintf(stderr, "p3d: unable to read pomPos poster\n");
      goto resetRef;
   }

   t3dInit(&mainToOrigin, T3D_EULER, T3D_ALLOW_CONVERSION);
   mainToOrigin.euler.euler[0] = pos.mainToOrigin.euler.yaw;
   mainToOrigin.euler.euler[1] = pos.mainToOrigin.euler.pitch;
   mainToOrigin.euler.euler[2] = pos.mainToOrigin.euler.roll;
   mainToOrigin.euler.euler[3] = pos.mainToOrigin.euler.x;
   mainToOrigin.euler.euler[4] = pos.mainToOrigin.euler.y;
   mainToOrigin.euler.euler[5] = pos.mainToOrigin.euler.z;

   /*--------------------------------------------------*/
   /* Copy the current robot position in SDI for display purposes */
   SDI_F->p3dPosRobotToOrigin.thetaRobot = pos.mainToOrigin.euler.yaw;
   SDI_F->p3dPosRobotToOrigin.pitchRobot = pos.mainToOrigin.euler.pitch;
   SDI_F->p3dPosRobotToOrigin.rollRobot = pos.mainToOrigin.euler.roll;
   SDI_F->p3dPosRobotToOrigin.xRobot = pos.mainToOrigin.euler.x;
   SDI_F->p3dPosRobotToOrigin.yRobot = pos.mainToOrigin.euler.y;
   SDI_F->p3dPosRobotToOrigin.zRobot = pos.mainToOrigin.euler.z;
   /*--------------------------------------------------*/

again:
   t3dInit(&goalToOrigin, T3D_EULER, T3D_ALLOW_CONVERSION);
   goalToOrigin.euler.euler[3] = SDI_F->wpList.wps[0].x;
   goalToOrigin.euler.euler[4] = SDI_F->wpList.wps[0].y;

   if (SDI_F->wpList.wps[0].dir == GENPOS_BACKWARD_MOTION) 
	   mainToOrigin.euler.euler[0] += M_PI;
   /* Compute goal to sensor for theta U-turn 
    *
    * We remove the 3D part of the attitude (pitch and roll)
    * and Z since we don't have the Z value for the goal */
   t3dCopyIn(&o2m, &mainToOrigin);
   o2m.euler.euler[1] = o2m.euler.euler[2] = 0;
   o2m.euler.euler[5] = 0;
   o2m.euler.flags |= T3D_ALLOW_CONVERSION;
   t3dInvert(&o2m);
   t3dCompIn(&g2m, &goalToOrigin, &o2m);
   t3dConvertTo(T3D_EULER, &g2m);
   theta = atan2(g2m.euler.euler[4], g2m.euler.euler[3]);

   /* distance to goal */
   SDI_F->goalDist.distance = sqrt(
	 g2m.euler.euler[3] * g2m.euler.euler[3] +
	 g2m.euler.euler[4] * g2m.euler.euler[4]);

#if VERBOSE
   fprintf(stderr, "p3d: robot position %f x %f, goal %f x %f, d = %f\n", 
	 pos.mainToOrigin.euler.x, pos.mainToOrigin.euler.y,
	 SDI_F->wpList.wps[0].x, SDI_F->wpList.wps[0].y, SDI_F->goalDist.distance);
#endif

   if (SDI_F->goalDist.distance <= SDI_F->arcsP.goalToal){
      SDI_F->movingStatus.status = P3D_GOAL_REACHED;
      printf("p3d: goal is reached, distance to goal %f less than %f\n",
            SDI_F->goalDist.distance, SDI_F->arcsP.goalToal);

      /* consume current way point */
      if (SDI_F->wpList.number > 0) {
	  SDI_F->wpList.number--;
	  for(i=1; i<=SDI_F->wpList.number; i++)
	      SDI_F->wpList.wps[i-1] = SDI_F->wpList.wps[i];
      }
      if (SDI_F->wpList.number == 0)
	  goto resetRef;
      else
	  goto again;
   }

   /* Start timer */
#ifdef HAS_POSIX_CLOCK
   if (clock_gettime(CLOCK_REALTIME, &tp0) != OK) {
      memset(&tp0, 0, sizeof(tp0));
   }
#else
   if (gettimeofday(&tv0, NULL) == -1) {
      memset(&tv0, 0, sizeof(tv0));
   }
#endif /* HAS_POSIX_CLOCK */

   /* Save various value, so we can replay the planification process */
   saved_v = sol->v;
   saved_w = sol->w;
   saved_average_time = sol->average_time;


   errno = 0;   /* to be sure to get the errno corresponding to p3dPlan */

   result = p3dPlan(planner, map, &mainToOrigin, &goalToOrigin, NULL, NULL, P3D_FULL, sol);

   /* Compute elapsed time */
#ifdef HAS_POSIX_CLOCK
   if (clock_gettime(CLOCK_REALTIME, &tp1) != OK) {
      printf("p3d: planification time = %ldms\n",
	    (tp1.tv_sec - tp0.tv_sec)*1000 + 
	    (tp1.tv_nsec - tp0.tv_nsec)/1000000);
   }
#else
   if (gettimeofday(&tv1, NULL) != -1) {
      printf("p3d: planification time = %ldms\n",
	    (tv1.tv_sec - tv0.tv_sec)*1000 + 
	    (tv1.tv_usec - tv0.tv_usec)/1000);
   }
#endif /* HAS_POSIX_CLOCK */
   if (SDI_F->dumpParams.status == P3D_DUMP_ON)
	   p3dDumpToFile(&mainToOrigin, saved_v, saved_w, saved_average_time);


   if (result) {
      fprintf(stderr,"p3d: plan failed\n");

	if(errno != 0)   /* No need to print "Success"... */
	   perror("p3d: plan");
	SDI_F->movingStatus.status = P3D_FAILED;

     goto resetRef;
   }

   /** Copy positions and corresponding configurations on the current selected arc in the SDI */
   solutionToPoster();

   /* Check wether or not we have a path to execute based on dMin */
   if (sol->d < SDI_F->arcsP.dMin && 
		   SDI_F->goalDist.distance >= SDI_F->arcsP.goalToal + SDI_F->arcsP.dMin) {
      fprintf(stderr, "p3d: distance %f m less than dMin (%f m)\n",
	    sol->d, SDI_F->arcsP.dMin);

      SDI_F->movingStatus.blocked_cycles++;
      if(SDI_F->movingStatus.blocked_cycles < 3)
	 fprintf(stderr, "p3d: blocked, trying again (%d)\n", SDI_F->movingStatus.blocked_cycles);
      else
	 SDI_F->movingStatus.status = P3D_BLOCKED;

      goto resetRef;
   }

   fprintf(stderr, "p3d: cost is %f\n", sol->cost);
   SDI_F->movingStatus.blocked_cycles = 0;

   SDI_F->speedRef.w = sol->w;
   SDI_F->speedRef.v = sol->v;

   if (SDI_F->goalDist.distance <= sol->v * EXEC_TASK_PERIOD(P3D_EXECTASK_NUM))
   {
	   double max_v = SDI_F->goalDist.distance / EXEC_TASK_PERIOD(P3D_EXECTASK_NUM);
	   double ratio = max_v / sol->v;

	   SDI_F->speedRef.v = max_v;
	   SDI_F->speedRef.w = sol->w * ratio;

	   fprintf(stderr, "p3d : Limit speed near goal from (%lf, %lf) to (%lf, %lf)\n",
					   sol->v, sol->w, SDI_F->speedRef.v, SDI_F->speedRef.w);
   }

   if (fabs(SDI_F->speedRef.w) > SDI_F->geomP.wMax) {

      if (fabs(sol->curv) < 1e-5) {
	 fprintf(stderr, "p3d: cannot do curv=%f\n", sol->curv);
	 SDI_F->movingStatus.status = P3D_FAILED;
	 goto resetRef;
      }
      SDI_F->speedRef.w = SDI_F->geomP.wMax * ((sol->curv > 0.0)?1:-1);
      SDI_F->speedRef.v = SDI_F->speedRef.w / sol->curv;

	  fprintf(stderr, "p3d: w is too high, going from (%lf,%lf) to (%lf,%lf)\n", 
			  sol->v, sol->w, SDI_F->speedRef.v, SDI_F->speedRef.w); 
      if (SDI_F->speedRef.v < 1e-3) {
	 fprintf(stderr, "p3d: too slow (%f m/s)\n", SDI_F->speedRef.v);
	 SDI_F->movingStatus.status = P3D_FAILED;
	 // A voir : etre moins categorique et faire un compteur, un certain nb d'erreurs "too slow" avant d'avoir l'echec
	 goto resetRef;
      }
   }

   if (isnan(SDI_F->speedRef.v) || isnan(SDI_F->speedRef.w)) {
      fprintf(stderr,"p3d: speed is nan\n");
      goto resetRef;
   }  


   /* send it */
   timeWait = sol->d / SDI_F->speedRef.v;
#if defined(ATRV) || defined(ATRV_JUNIOR)
   SDI_F->speedRef.numRef++;
   SDI_F->speedRef.updatePeriod = timeWait / RFLEX_SEC_PERIOD;
#elif defined(RMP400)
   SDI_F->speedRef.numRef++;
   SDI_F->speedRef.updatePeriod = timeWait / RMP400_SEC_PERIOD;
#else
   SDI_F->speedRef.num++;
   SDI_F->speedRef.updatePeriod = timeWait / LLOCO_SEC_PERIOD;
#endif

   /* compute sleep time */
   nPeriods = 
      SDI_F->planP.period * timeWait/EXEC_TASK_PERIOD(P3D_EXECTASK_NUM);

   if (SDI_F->wpList.wps[0].dir == GENPOS_BACKWARD_MOTION) 
	   SDI_F->speedRef.v *= -1.0;

   /* inform user */
   fprintf(stderr,
		   "p3d: executing v=%.3f m/s w=%.3f rad/s r=%.3f m d=%.3f m\n",
		   SDI_F->speedRef.v, SDI_F->speedRef.w, 
		   (fabs(sol->curv) < 1e-3 ? 0.0f : 1/sol->curv), sol->d);
   fprintf(stderr, "p3d: waiting %f s (%d periods)\n", SDI_F->planP.period * timeWait, nPeriods);

   return OK;

resetRef:
   /* stop until next way point  */
   SDI_F->speedRef.v = SDI_F->speedRef.w = 0.0;
   SDI_F->speedRef.updatePeriod = 0;
   nPeriods = 0;
   return OK;
}







/* p3dPlaceRobotOnMapExec  -  codel EXEC de PlaceRobotOnMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dPlaceRobotOnMapExec (P3D_PLACEROBOT_PARAMS *placeRobotPos, int *bilan)
{
  P3DRobotConfig  config;

  if( p3dPlaceRobotOnMap(&config,  /* output : predicted robot config. */
			 planner, 
			 placeRobotPos->x, placeRobotPos->y, 
			 placeRobotPos->theta,
			 map, placeRobotPos->gdheDisplay) != 0 )
    {
      *bilan = S_p3d_BAD_PARAMS;
      fputs("p3dPlaceRobotOnMapExec: Failed to place Robot on map\n", stderr);
      return ETHER;
    }
    
  return ETHER;
}





/*------- Read demP3dPoster to get the p3dMap --------------*/
short
p3dReadDemMap (char*     demP3dPosterName,
	    P3D_MAP  *p3dMap )
{
  DTM_P3D_POSTER  demP3dPoster;
  short  bilan=0;
  
  
  /* Read demP3dPoster */
  if (!demP3dPosterId) {
     if (posterFind(demP3dPosterName, &demP3dPosterId) == ERROR) {
       fprintf(stderr, "p3d (readDemMap): cannot find %s dem poster\n", demP3dPosterName);
       return(bilan++); /* goto resetRef; */
    }
  }
  if (p3dDTM_P3D_POSTERPosterRead(demP3dPosterId, &(demP3dPoster)) == ERROR) {
    fprintf(stderr, "p3d (readDemMap): unable to read %s dem poster\n", demP3dPosterName);
    return(bilan++);  /* goto resetRef; */
  }

  /*-- Fill the P3D_MAP with info from the demPoster */
  if( demP3dPoster_to_p3dMap(&demP3dPoster, p3dMap)  )
    {
      SDI_F->movingStatus.status = P3D_FAILED_POSTER;
      return(bilan++);
    }


  return bilan;
}


/**************************************************************/
/****   MONITOR TASK      ************************************/


/*------------------------------------------------------------------------
 *
 * p3dMonitorTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
p3dMonitorTaskInit(int *report)
{

  SDI_F->monitorData.status = P3D_MONITOR_OFF;
  SDI_F->monitorParams.logFlag = P3D_MONITOR_LOG_OFF;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * p3dMonitorTaskPerm  --  codel of permanent activity
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */

STATUS
p3dMonitorTaskPerm(int *report)
{
  /* Nothing special to be done here for the moment ? */

  return OK;
}


/*------------------------------------------------------------------------
 * ConfigMonitor
 *
 * Description: 
 *
 * Reports:      OK
 *              S_p3d_PLACEMENT_FAILED
 */

/* p3dConfigMonitorStart  -  codel START of ConfigMonitor
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dConfigMonitorStart(P3D_MONITOR_PARAMS *monitorParams, int *report)
{

  /* Look for pom Poster ID (Already done in Init Request) */
/*   if (posterFind(SDI_F->initParams.pomPoster, &pomPosterId) == ERROR) { */
/* 	fprintf(stderr, "p3d: cannot find pom poster\n"); */
/* 	*report = S_p3d_POSTER_NOT_FOUND; */
/* 	return ETHER; */
/*     } */

  SDI_F->monitorData.status = P3D_MONITOR_ON;

  /*----------- Save Logs on file if asked -------------*/
  if(monitorParams->logFlag == P3D_MONITOR_LOG_ON) {
    file = fopen(monitorParams->fileName, "w");
    if (file == NULL) {
      fprintf(stderr,"p3dPredictedConfigMonitorStart: Creation of log file Failed\n");
      return END;
    }

    fprintf(stderr, "p3dPredictedConfigMonitorStart: Logging Monitor Data in file %s \n", monitorParams->fileName);

    /* Write a first line in file containing the names of signals logged*/
    fprintf(stderr, "p3dPredictedConfigMonitorStart: Writing Head of Log File...");
    fprintf(file, "%s\t %s\t%s\t%s\t %s\t%s\t%s\t%s\t %s\t%s\t %s\n",
	    "time",
	    "xRobot", "yRobot", "thetaRobot",
	    "obsPitch", "obsRoll",
	    "predPitch", "predRoll",
	    "probaPitch", "probaRoll", 
	    "behavior" );
    fprintf(stderr, " Done\n");
  }

  return EXEC;
}


/* p3dConfigMonitorExec  -  codel EXEC of ConfigMonitor
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dConfigMonitorExec(P3D_MONITOR_PARAMS *monitorParams, int *report)
{
  POM_POS pospom;
  double xRobot, yRobot, zRobot;
  double thetaRobot, pitchRobot, rollRobot;

  /*-- Time --*/
#ifdef HAS_POSIX_CLOCK
  struct timespec tp0, tp1;
#else
  struct timeval tv0, tv1;
#endif /* HAS_POSIX_CLOCK */
  static double monitorTime;  /* current time (sec) */
  /*-------------------------------------*/

  /* Get Current Position */
  if (p3dPOM_POSPosterRead(pomPosterId, &(pospom)) == ERROR) {
    fprintf(stderr, "p3dConfigMonitor: unable to read pomPos poster\n");
    return ERROR;
  }

  /*------------- Start timer ------------------------------*/
#ifdef HAS_POSIX_CLOCK
   if (clock_gettime(CLOCK_REALTIME, &tp0) != OK) {
      memset(&tp0, 0, sizeof(tp0));
   }
#else
   if (gettimeofday(&tv0, NULL) == -1) {
      memset(&tv0, 0, sizeof(tv0));
   }
#endif /* HAS_POSIX_CLOCK */
  /*---------- End start time -------------------------------*/

  xRobot = pospom.mainToOrigin.euler.x;
  yRobot = pospom.mainToOrigin.euler.y;
  zRobot = pospom.mainToOrigin.euler.z;
  thetaRobot = pospom.mainToOrigin.euler.yaw;
  pitchRobot = pospom.mainToOrigin.euler.pitch;
  rollRobot  = pospom.mainToOrigin.euler.roll; /* TO BE CHECKED (asin?) */
 
  /* Compute predicted config on current location */
  /* NB: the map used here is the one read in the main ExecTask */
  /*  to avoid multiple posters read */
  if( p3dPlaceRobotOnMap(&predRobotConfig,  /* output : predicted robot config. */
			 planner, 
			 xRobot, yRobot, RAD_TO_DEG(thetaRobot), /* A verifier */
			 map, 2) != 0)  /* Display should be == 0 ?? */
    {
      *report = S_p3d_PLACEMENT_FAILED;
      fputs("p3dPredictedConfigMonitorExec: Failed to place Robot on map\n", stderr);
      return END;
    }

  SDI_F->monitorData.distPitch = predRobotConfig.pitch - pitchRobot;
  SDI_F->monitorData.distRoll  = predRobotConfig.roll  - rollRobot;

  /* Get the "shaped" dist. obs-pred (-> "behavior pseudo-proba") */
  SDI_F->monitorData.probaPitch = p3dShapeDistrib(
				    SDI_F->monitorData.distPitch,
				    planner->geom->pitchMax/2.0,
				    pitchRobot, planner->geom->pitchMax,
				    predRobotConfig.sigmaPitch);
  SDI_F->monitorData.probaRoll  = p3dShapeDistrib(
				    SDI_F->monitorData.distRoll,
				    planner->geom->rollMax/2.0,
				    rollRobot, planner->geom->rollMax,
				    predRobotConfig.sigmaRoll);
  /* THE outuput */
  SDI_F->monitorData.behavior = MAX(SDI_F->monitorData.probaPitch,
				    SDI_F->monitorData.probaRoll);
  /* data exported in monitor poster */


  /*------------- Compute elapsed time --------------------------*/
  /* Compute elapsed time */
#ifdef HAS_POSIX_CLOCK
   if (clock_gettime(CLOCK_REALTIME, &tp1) != OK) {
      printf("p3d: monitoring time = %ldms\n",
	     (tp1.tv_sec - tp0.tv_sec)*1000 + 
	     (tp1.tv_nsec - tp0.tv_nsec)/1000000);
      /* current time (sec) */
      monitorTime = ((double) tp1.tv_sec) + ((double) tp1.tv_nsec)/1000000000.0;
   }
#else
   if (gettimeofday(&tv1, NULL) != -1) {
     printf("p3d: monitoring time = %ldms\n",
	    (tv1.tv_sec - tv0.tv_sec)*1000 + 
	    (tv1.tv_usec - tv0.tv_usec)/1000);
      /* current time (sec) */
     monitorTime = ((double) tv1.tv_sec) + ((double) tv1.tv_usec)/1000000.0;
   }
#endif /* HAS_POSIX_CLOCK */
  /*------------- End elapsed time ------------------------------*/


#if VERBOSE
  /*----- Print some results info ----------*/
  fprintf(stderr, "p3dMonitor: ----- Predicted Config. -----\n");
  fprintf(stderr,"p3dMonitor: roll(deg): %lf pitch(deg) = %lf\n", 
	  RAD_TO_DEG(predRobotConfig.roll), RAD_TO_DEG(predRobotConfig.pitch) );
  fprintf(stderr,"p3dMonitor: alpha1(deg): %lf\n", 
	  RAD_TO_DEG(predRobotConfig.alpha1) );
  fprintf(stderr,"p3dMonitor: dzBodyGround: %lf\n", predRobotConfig.dzBodyGround);
  
  fprintf(stderr, "p3dMonitor: ----- Observed Config. -----\n");
  fprintf(stderr, "p3dMonitor: Position: x=%lf, y=%lf, z=%lf, theta=%lf\n", 
	  xRobot, yRobot, zRobot, RAD_TO_DEG(thetaRobot) );
  fprintf(stderr,"p3dMonitor: roll(deg): %lf pitch(deg) = %lf\n", 
	  RAD_TO_DEG(rollRobot), RAD_TO_DEG(pitchRobot) );
  fprintf(stderr, "----- Difference: -----\n");
  fprintf(stderr, "p3dMonitor: diffPitch(deg) = %lf, diffRoll(deg) = %lf\n",
	  RAD_TO_DEG(predRobotConfig.pitch - pitchRobot),
	  RAD_TO_DEG(predRobotConfig.roll - rollRobot) );
  fprintf(stderr, "----- Probas: -----\n");
  fprintf(stderr, "p3dMonitor: probaPitch = %lf, probaRoll = %lf\n", 
	  SDI_F->monitorData.probaPitch, SDI_F->monitorData.probaRoll);
  fprintf(stderr, "p3dMonitor: BEHAVIOR = %lf\n", 
	  SDI_F->monitorData.behavior);
  fprintf(stderr, "----------------------------------------\n\n");
#endif /* VERBOSE */

  /*----------- Save Logs on file if asked -------------*/
  if(monitorParams->logFlag == P3D_MONITOR_LOG_ON) {
    fprintf(file, "%lf\t %lf\t%lf\t%lf\t %lf\t%lf\t%lf\t%lf\t %lf\t%lf\t %lf\n",
	    monitorTime,
	    xRobot, yRobot, thetaRobot,
	    pitchRobot, rollRobot,
	    predRobotConfig.pitch, predRobotConfig.roll,
	    SDI_F->monitorData.probaPitch, SDI_F->monitorData.probaRoll, 
	    SDI_F->monitorData.behavior );
  }

  return EXEC;
}

/* p3dConfigMonitorEnd  -  codel END of ConfigMonitor
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
p3dConfigMonitorEnd(P3D_MONITOR_PARAMS *monitorParams, int *report)
{
  
  if(monitorParams->logFlag == P3D_MONITOR_LOG_ON) {
    fclose(file);
    fprintf(stderr, "p3dPredictedConfigMonitorEnd: Log file %s closed\n", monitorParams->fileName);
    SDI_F->monitorParams.logFlag = P3D_MONITOR_LOG_OFF;    
  }

  SDI_F->monitorData.status = P3D_MONITOR_OFF;

  return ETHER;
}

static int
p3dDumpEvaluateNbArcs(int n, int depth)
{
	if (depth <= 0)
		return 0;

	int nArcs_currentlevel;
	int nR = n / 2;	
	nArcs_currentlevel = nR * 2 + 1;

	return nArcs_currentlevel * (1 + p3dDumpEvaluateNbArcs(2* nR/2 + 1, depth -1));
}

static void
p3dDumpToFile(T3D* m2o, double saved_v, double saved_w, double average_time)
{
	char buf[512];
	FILE * f;
	int max_arcs;
	int i, j;
	struct timeval tv;

	/* Update the content of ArcsDump */
	t3dConvertTo(T3D_EULER, m2o);
	t3dConvertTo(T3D_EULER, &goalToOrigin);
	p3dDumpP3dArcs(planner->arcs, m2o, SDI_F->arcsP.nArcs, ArcsDump);
	snprintf(buf, sizeof(buf), "%s/p3d_map.%04d", SDI_F->dumpParams.baseName, SDI_F->dumpParams.num);
	write_p3dmap(map, buf);

	snprintf(buf, sizeof(buf), "%s/arc_dump.%04d.txt", SDI_F->dumpParams.baseName, SDI_F->dumpParams.num);
	SDI_F->dumpParams.num++;

	f = fopen(buf, "w+");
	if (f == NULL) {
		fprintf(stderr, "Can't open %s : %s\n", buf, strerror(errno));
		return;
	}

	fprintf(f, "%lf %lf %lf\n", m2o->euler.euler[3], m2o->euler.euler[4], m2o->euler.euler[0]);
	fprintf(f, "%lf %lf\n", goalToOrigin.euler.euler[3], goalToOrigin.euler.euler[4]);
	fprintf(f, "%f %f %f\n", saved_v, saved_w, average_time);
	gettimeofday(&tv, NULL);
	fprintf(f, "%d:%d\n", tv.tv_sec, tv.tv_usec);
	fprintf(f, "%lf %lf\n", sol->v, sol->w);

	max_arcs = ArcsDump[0].nArcsTotal;
	for (i = 0; i < max_arcs; i++) {
		int max_nodes = ArcsDump[i].nNodes; 
		if (ArcsDump[i].is_elected) {
			fprintf(f, "Arc %d is elected\n", i);
		}
		fprintf(f, "Arc %d r : %lf\n", i, ArcsDump[i].r);
		fprintf(f, "Arc %d cost : %lf\n", i, ArcsDump[i].cost);
		fprintf(f, "Arc %d cost_balanced : %lf\n", i, ArcsDump[i].cost_balanced);
		fprintf(f, "Arc %d current : %d\n", i, ArcsDump[i].current);
		fprintf(f, "Arc %d first_unknow : %d\n", i, ArcsDump[i].first_unknow);
		for (j = 0; j < max_nodes; j++) {
			struct P3DNodeDump node = ArcsDump[i].nodes[j];
			fprintf(f, "Coords of node %d of arc %d are [ %lf %lf %lf %lf %lf %lf ]\n",j, i,
					node.x, node.y, node.z, node.yaw, node.pitch, node.roll);
			fprintf(f, "Cost of node %d of arc %d is %lf\n", j, i, node.cost);
			fprintf(f, "Dubins of node %d of arc %d is %lf\n", j, i, node.remain);
			fprintf(f, "Global cost of node %d of arc %d is %lf\n", j, i, node.cost + node.remain);
		}
	}

	fclose(f);
}

