/**
 ** demP3dPosterToP3dMap.c
 **
 ** Functions for conversion from demP3dPoster to p3dMap (calife struct)
 **
 ** Auteur: Thierry Peynot
 ** Date: June 8, 2005
 **
 **/

#ifndef UNIX
#include <vxWorks.h>
#include <timers.h>
#else
#include <portLib.h>
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "server/p3dHeader.h"

#include <t3d/t3d.h>

#include "dtmStruct.h"
#include "demP3dPosterToP3dMap.h"




/*--- Fill a P3D_MAP with info read in a demP3dPoster --------*/
short
demP3dPoster_to_p3dMap(DTM_P3D_POSTER *demP3dPoster,
		       P3D_MAP        *p3dMap )
{
  short  bilan=0;
  int    i,j;
  

  if(demP3dPoster == NULL)
    {
      printf("demP3dPoster_to_p3dMap: error: no demP3dPoster");
      return(bilan++);
    }
  
  /* Check that the demMap dimensions are <= dimensions of P3D_MAP */
  if( ( demP3dPoster->nbLines > P3D_MAP_MAX_LINES) || (demP3dPoster->nbCols > P3D_MAP_MAX_COLUMNS) )
    {
      fputs("p3d: Error: demPoster size too big!! \n", stderr);
      return(bilan++);
    }

  /* Now, Copy map info */
  p3dMap->nbLines = demP3dPoster->nbLines;
  p3dMap->nbCols = demP3dPoster->nbCols;
  p3dMap->xOrigin = demP3dPoster->xOrigin;
  p3dMap->yOrigin = demP3dPoster->yOrigin;
  
  /* Create the T3D mainToOrigin using the xOrigin and yOrigin */
  p3dMap->mainToOrigin->type = T3D_EULER;
  p3dMap->mainToOrigin->euler.euler[0] = 0.0;
  p3dMap->mainToOrigin->euler.euler[1] = 0.0;
  p3dMap->mainToOrigin->euler.euler[2] = 0.0;
  p3dMap->mainToOrigin->euler.euler[3] = demP3dPoster->xOrigin;
  p3dMap->mainToOrigin->euler.euler[4] = demP3dPoster->yOrigin;
  p3dMap->mainToOrigin->euler.euler[5] = 0.0;


  /* xScale and yScale should be the same */
  if( demP3dPoster->xScale == demP3dPoster->yScale )
    p3dMap->scale = demP3dPoster->xScale;
  else
    {
      fputs("p3d: Error in demP3dPoster: xScale != yScale \n", stderr);
      return(bilan++);
    }
  
  p3dMap->zOrigin = demP3dPoster->zOrigin;
  p3dMap->zScale   = demP3dPoster->zScale;
  

  /*-- Copy cells --*/
  for(i = 0; i < demP3dPoster->nbLines; i++)
    {
      for(j = 0; j < demP3dPoster->nbCols; j++)
	{
	  /* altitude (cf dem module for the transfo needed) */
	  /* Should put 0 if cell is empty */
	  if(demP3dPoster->state[i][j] == DTM_CELL_FILLED)
	    {
	      p3dMap->cells[i][j].z = demP3dPoster->zfloat[i][j];
	      /* cell state */
	      p3dMap->cells[i][j].state = P3D_MAP_CELL_FILLED;
	    }
	  else /* cell is empty */
	    {
	      p3dMap->cells[i][j].z = 0.0;
	      /* cell state */
	      p3dMap->cells[i][j].state = P3D_MAP_CELL_EMPTY;
	    }
	  
/* 	  if(p3dMap->cells[i][j].state == P3D_MAP_CELL_FILLED) */
/* 	    printf("p3dmap_cell[%d][%d] is FILLED: %d\n", i, j, p3dMap->cells[i][j].state); */
/* 	  if(demP3dPoster->state[i][j] == DTM_CELL_FILLED) */
/* 	    printf("demP3dposter_cell[%d][%d] is FILLED: %d\n", i, j, demP3dPoster->state[i][j]); */	  
	}
    }

 return(bilan);
}
